/*
 * Copyright CERN 2013
 * Author: Federico Vaga <federico.vaga@gmail.com>
 */
#ifndef ADC_LIB_INT_H_
#define ADC_LIB_INT_H_

/*
 * offsetof and container_of come from kernel.h header file
 */
#define ARRAY_SIZE(x) (sizeof(x) / sizeof(x[0]))
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#define container_of(ptr, type, member) ({			\
	const typeof( ((type *)0)->member ) *__mptr = ((void *)ptr);	\
	(type *)( (char *)__mptr - offsetof(type,member) );})
#define to_dev_zio(dev) (container_of(dev, struct __adc_dev_zio, gid))

/* ->open takes different args than open(), so fa a fun to use tpyeof */
struct adc_board_type;
struct adc_dev *adc_internal_open(const struct adc_board_type *b,
					unsigned int dev_id,
					unsigned long totalsamples,
					unsigned int nbuffer,
					unsigned long flags);

/**
 * The operations structure is the device-specific backend of the library
 */
struct adc_operations {
	typeof(adc_internal_open)	*open; /**< @related adc_open */
	typeof(adc_close)		*close; /**< @related adc_close */

	typeof(adc_acq_start)		*acq_start; /**< @related adc_acq_start */
	typeof(adc_acq_poll)		*acq_poll; /**< @related adc_acq_poll */
	typeof(adc_acq_stop)		*acq_stop; /**< @related adc_acq_stop */

	typeof(adc_apply_config)	*apply_config; /**< @related adc_apply_config */
	typeof(adc_retrieve_config)	*retrieve_config; /**< @related adc_retrieve_config */
	typeof(adc_get_param)		*get_param; /**< @related adc_acq_get_param */
	typeof(adc_set_param)		*set_param; /**< @related adc_acq_set_param */

	typeof(adc_request_buffer)	*request_buffer; /**< @related adc_request_buffer */
	typeof(adc_fill_buffer)	*fill_buffer; /**< @related adc_fill_buffer */
	typeof(adc_tstamp_buffer)	*tstamp_buffer; /**< @related adc_tstamp_buffer */
	typeof(adc_release_buffer)	*release_buffer; /**< @related adc_release_buffer */

	typeof(adc_trigger_fire)	*trigger_fire; /**< @related adc_trigger_fire */
};


/**
 * This structure describes the board supported by the library
 */
struct adc_board_type {
	char *name; /**< name of the board type, for example "fmc-adc-100MS" */
	char *devname; /**< name of the device in Linux */
	char *driver_type; /**< the kind of driver that hanlde this kind of
			      board (e.g. ZIO) */

	uint32_t capabilities[__ADC_CONF_TYPE_LAST_INDEX]; /**< bitmask of
							      device
							      capabilities */
	struct adc_operations *adc_op; /**< pointer to a set of operations */
};

/**
 * Generic Instance Descriptor
 */
struct adc_gid {
	const struct adc_board_type *board; /**< board */
};

/* Definition of board types */
extern struct adc_board_type fmcadc_100ms_4ch_14bit;
extern struct adc_board_type adc_ziofake;
extern struct adc_board_type adc_genericfake;

/**
 * Internal structure for ZIO drivers only
 */
struct __adc_dev_zio {
	unsigned int cset; /**< channel-set index */
	int fdc; /**< control file descriptor */
	int fdd; /**< data file descriptor */
	uint32_t dev_id; /**< ZIO device ID */
	unsigned long flags; /**< internal flags */
	char *devbase; /**< base path to char-device */
	char *sysbase; /**< base path to device sysfs */
	unsigned long samplesize; /**< size of 1 sample */
	unsigned long pagesize; /**< size of 1 page */
	/* Mandatory field */
	struct adc_gid gid; /**< general ADC descriptor */
};

/**
 * Configuration options for the generic-fake board
 */
struct __adc_genfake_conf {
	int nshots; /**< number of shots  */
	int presamples; /**< number of pre-samples */
	int postsamples; /**< number of post-samples */
};

/**
 * Internal structure for generic-fake board type
 */
struct __adc_dev_genfake {
	int acq; /**< acquisition status */
	unsigned int conf_index; /**< conf_index */
	unsigned int conf_type; /**< conf_type */
	unsigned long samplesize; /**< size of 1 sample */
	struct __adc_genfake_conf *cf; /**< local configuration */
	struct adc_gid gid; /**< general ADC descriptor */
};

#define ADC_FLAG_VERBOSE 0x00000001 /**< Internal verbose mode */
#define ADC_FLAG_MALLOC  0x00000002 /**< allocate data with malloc */
#define ADC_FLAG_MMAP    0x00000004 /**< use mmap to get data */

/* The board-specific functions are defined in fmc-adc-100m14b4cha.c */
#define ADC_NAME_FMCADC100M "fmc-adc-100m14b4cha"
extern struct adc_board_type fmcadc_100ms_4ch_14bit;

#define ADC_NAME_GENERICFAKE "adc-genericfake"
extern struct adc_board_type adc_genericfake;

#define ADC_NAME_ZIOFAKE "adc-zzero"
extern struct adc_board_type adc_ziofake;


/* ZIO boards */

struct adc_dev *adc_zio_open(const struct adc_board_type *b,
				   unsigned int dev_id,
				   unsigned long totalsamples,
				   unsigned int nbuffer,
				   unsigned long flags);

int adc_zio_close(struct adc_dev *dev);

int adc_zio_acq_start(struct adc_dev *dev,
			unsigned int flags,
			struct timeval *timeout);

int adc_zio_acq_poll(struct adc_dev *dev,
			unsigned int flags,
			struct timeval *timeout);

int adc_zio_acq_stop(struct adc_dev *dev, unsigned int flags);

struct adc_buffer *adc_zio_request_buffer(struct adc_dev *dev,
						int nsamples,
						void *(*alloc)(size_t),
						unsigned int flags);

int adc_zio_fill_buffer(struct adc_dev *dev,
			   struct adc_buffer *buf,
			   unsigned int flags,
			   struct timeval *timeout);

struct adc_timestamp *adc_zio_tstamp_buffer(struct adc_buffer *buf,
						  struct adc_timestamp *);

int adc_zio_release_buffer(struct adc_dev *dev,
			      struct adc_buffer *buf,
			      void (*free_fn)(void *));

		/* The following functions are in config-zio.c */
int adc_zio_apply_config(struct adc_dev *dev, unsigned int flags,
			    struct adc_conf *conf);

int adc_zio_retrieve_config(struct adc_dev *dev,
			       struct adc_conf *conf);

int adc_zio_set_param(struct adc_dev *dev, char *name,
			 char *sptr, int *iptr);

int adc_zio_get_param(struct adc_dev *dev, char *name,
			 char *sptr, int *iptr);

int adc_zio_sysfs_set(struct __adc_dev_zio *fa, char *name,
		     uint32_t *value);


/*adc-genericfake*/

#endif /* ADC_LIB_INT_H_ */
