/*
 * Copyright CERN 2013
 * Author: Federico Vaga <federico.vaga@gmail.com>
 */

#ifndef ADC_LIB_H_
#define ADC_LIB_H_

#ifdef __cplusplus
#pragma GCC diagnostic ignored "-Wwrite-strings"
extern "C" {
#endif

#include <stdint.h>
#include <stdlib.h>
#include <sys/time.h>

/* Error codes start from 1024 to void conflicting with libc codes */
#define __ADC_ERRNO_START 1024
#define ADC_ENOP		1024
#define ADC_ENOCAP		1025
#define ADC_ENOCFG		1026
#define ADC_ENOGET		1027
#define ADC_ENOSET		1028
#define ADC_ENOCHAN		1029
#define ADC_ENOMASK		1030
#define ADC_EDISABLED		1031
#define ADC_EROUTE		1032

/**
 * Opaque type. any instance of this should be used as token
 * to use the API
 */
struct adc_dev;

extern const char *adc_board_name[];

/**
 * Enumerate all supported boards
 */
enum adc_supported_boards {
	FMCADC_100MS_4CH_14BIT = 0, /**< It identifies the FMC ADC 100M 14Bit
				       4Channel */
	ADC_ZIOFAKE, /**< It identifies a generic board that supports the ZIO
			framework */
	ADC_GENERICFAKE, /**< It identifies a generic board for testing
			    purpose */
	__ADC_SUPPORTED_BOARDS_LAST_INDEX, /**< It represents the the last
					      index of enum. It can be useful
					      for some sort of automation */
};


/**
 * The buffer hosts data and metadata, plus informative fields
 */
struct adc_buffer {
	void *data; /**< container for all the samples */
	void *metadata; /**< description of the acquisition */
	int samplesize; /**< size of a single sample in Byte (a sample for
			   each channel, e.g 4 chan 14bit sample => samplesize = 8 Byte) */
	int nsamples; /**< number of samples in the buffer */
	struct adc_dev *dev; /**< the device used for the acquisition */
	void *mapaddr; /**< mmap address */
	unsigned long maplen; /**<  mmap length */
	unsigned long flags; /**< internal to the library */
};


/**
 * Timestamp descriptor. This is should be big enough to store *any*
 * time stamp.
 */
struct adc_timestamp {
	uint64_t secs; /**< number of seconds */
	uint64_t ticks; /**< number of coarse sub-seconds unit */
	uint64_t bins; /**< high resolution sub-seconds unit */
};


/**
 * It describes the possible polarity values
 */
enum adc_trigger_polarity {
	ADC_TRG_POL_POS = 0, /**< positive edge/slope */
	ADC_TRG_POL_NEG, /**< negative edge/slope */
};

/**
 * It describes the possible configuration option for an external trigger
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_trigger_ext {
	ADC_CONF_TRG_EXT_ENABLE, /**< It enable/disable the trigger */
	ADC_CONF_TRG_EXT_POLARITY, /**< It is used to apply the polarity to the
				      external triggers */
	ADC_CONF_TRG_EXT_DELAY, /**< acquisition delay after trigger */
	__ADC_CONF_TRG_EXT_LAST_INDEX, /**< It represents the the last index of
					  this enum. It can be useful for some
					  sort of automation */
};


/**
 * It describes the possible configuration option for a threshold trigger
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_trigger_thr {
	ADC_CONF_TRG_THR_ENABLE, /**< It enable (1) or disable (0) the trigger */
	ADC_CONF_TRG_THR_POLARITY, /**< It is used to apply the polarity to the
				      threshold triggers */
	ADC_CONF_TRG_THR_DELAY, /**< It defines the acquisition delay after
				   trigger */
	ADC_CONF_TRG_THR_THRESHOLD, /**< The threshold value that triggers
				       the acquisition */
	ADC_CONF_TRG_THR_HYSTERESIS, /**< If defines the hysteresis associated
					to the threshold value */
	__ADC_CONF_TRG_THR_LAST_INDEX, /**< It represents the the last index
					  of this enum. It can be useful for
					  some sort of automation */
};


/**
 * It describes the possible configuration option for a time trigger
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_trigger_tim {
	ADC_CONF_TRG_TIM_ENABLE, /**< It enable (1) or disable (0) the trigger */
	__ADC_CONF_TRG_TIM_LAST_INDEX, /**< It represents the the last index
					  of this enum. It can be useful for
					  some sort of automation */
};


/**
 * This represents the list of configuration options for the acquisition
 * when using ``conf->type = ADC_CONF_TYPE_ACQ``
 */
enum adc_configuration_acquisition {
	ADC_CONF_ACQ_N_SHOTS = 0, /**< umber of shots during the acquisition */
	ADC_CONF_ACQ_POST_SAMP, /**< number of samples after the trigger */
	ADC_CONF_ACQ_PRE_SAMP, /**<number of sample before the trigger  */
	ADC_CONF_ACQ_UNDERSAMPLE, /**< Acquisition undersample factor */
	ADC_CONF_ACQ_FREQ_HZ, /**< Acquisition frequency */
	ADC_CONF_ACQ_N_BITS, /**< Number of bit per sample */
	__ADC_CONF_ACQ_ATTRIBUTE_LAST_INDEX, /**< It represents the the last
						index of this enum. It can be
						useful for some sort of
						automation */
};


/**
 * This represents the list of configuration options for the channels
 * when using ``conf->type = ADC_CONF_TYPE_CHN``
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_channel {
	ADC_CONF_CHN_RANGE = 0, /**< Volt range */
	ADC_CONF_CHN_TERMINATION, /**< channel Termination */
	ADC_CONF_CHN_OFFSET, /**< signal offset */
	ADC_CONF_CHN_SATURATION, /**< saturation value */
	ADC_CONF_CHN_GAIN, /**< signal gain */
	__ADC_CONF_CHN_ATTRIBUTE_LAST_INDEX, /**< It represents the the last
						index of this enum. It can be
						useful for some sort of
						automation */
};


/**
 * This represents the list of configuration options for the board status
 * when using ``conf->type = ADC_CONF_TYPE_BRD``
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_board {
	ADC_CONF_BRD_STATUS = 0, /**< Board status */
	ADC_CONF_BRD_MAX_FREQ_HZ, /**< Maximum frequency in HZ */
	ADC_CONF_BRD_MIN_FREQ_HZ, /**< Minimum frequency in HZ */
	ADC_CONF_BRD_STATE_MACHINE_STATUS, /**< Acquisition state machine status */
	ADC_CONF_BRD_N_CHAN, /**< Number of acquisition channel */
	ADC_CONF_BRD_N_TRG_EXT, /**< Number of external triggers */
	ADC_CONF_BRD_N_TRG_THR, /**< Number of threshold triggers */
	ADC_CONF_BRD_N_TRG_TIM, /**< Number of timer triggers */
	/* TODO seconds, ticks and bins are bit -> High - low */
	ADC_CONF_UTC_TIMING_BASE_S,  /**< Board internal time: seconds */
	ADC_CONF_UTC_TIMING_BASE_T,  /**< Board internal time: coarse
					sub-seconds */
	ADC_CONF_UTC_TIMING_BASE_B,  /**< Board internal time: fine
					sub-seconds*/
	__ADC_CONF_BRD_ATTRIBUTE_LAST_INDEX, /**< It represents the the last
						index of this enum. It can be
						useful for some sort of
						automation */
};


/**
 * It describes the possible configuration types
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_type {
	ADC_CONF_TYPE_BRD = 0, /**< Configuration for the board */
	ADC_CONF_TYPE_CHN, /**< Configuration for an acquisition channels */
	ADC_CONF_TYPE_ACQ, /**< Configuration for the acquisition */
	ADC_CONF_TYPE_CUS, /**< Custom configuration for board-specific
			      options for any of the possible domain: board,
			      channel, triggers. */
	ADC_CONF_TYPE_TRG_EXT, /**< Configuration for external triggers */
	ADC_CONF_TYPE_TRG_THR, /**< Configuration for threshold triggers */
	ADC_CONF_TYPE_TRG_TIM, /**< Configuration for time triggers */
	__ADC_CONF_TYPE_LAST_INDEX, /**< It represents the the last index of
				       this enum. It can be useful for some
				       sort of automation */
};


#define __ADC_CONF_LEN 64 /* number of allocated items in each structure */

/**
 * ADC configuration descriptor.
 */
struct adc_conf {
	enum adc_configuration_type type; /**< configuration type */
	uint32_t dev_type; /**< device type */
	uint32_t route_to; /**< internal route to a particular sub-device
			      (e.g. a channel) */
	uint32_t flags; /**< FIXME how to identify invalid? */
	uint64_t mask; /**< capabilities mask, 1 when the correspondent
			  value in ``value[]`` is valid */
	uint32_t value[__ADC_CONF_LEN]; /**< array of configuration value.
					   The position in the array correspond
					   to a specific capability */
};


#define ADC_F_USERMASK	0xffff0000 /**< Flag mask -- low-bits are used
				      by lib-int.h */
#define ADC_F_FLUSH	0x00010000 /**< Flag used to flush the buffer
				      (used by adc_open) */
#define ADC_F_VERBOSE	0x00020000 /**< Flag used to verbose on stdout/stderr
				      (usable by any function)*/

/**
 * @defgroup dev Basic
 * Basic library functions
 * @{
 */
extern void adc_print_version();
extern int adc_init(void);
extern void adc_exit(void);
extern char *adc_strerror(int errnum);
extern char *adc_get_driver_type(struct adc_dev *dev);
extern struct adc_dev *adc_open(char *name, unsigned int dev_id,
				      unsigned long totalsamples,
				      unsigned int nbuffer,
				      unsigned long flags);
extern struct adc_dev *adc_open_by_lun(char *name, int lun,
					     unsigned long totalsamples,
					     unsigned int nbuffer,
					     unsigned long flags);
extern int adc_close(struct adc_dev *dev);
extern int adc_trigger_fire(struct adc_dev *dev);
extern int adc_has_trigger_fire(struct adc_dev *dev);
/**@}*/


/**
 * @defgroup acq Acquisition
 * Acquisition functions
 * @{
 */
extern int adc_acq_start(struct adc_dev *dev, unsigned int flags,
			    struct timeval *timeout);
extern int adc_acq_poll(struct adc_dev *dev, unsigned int flags,
			    struct timeval *timeout);
extern int adc_acq_stop(struct adc_dev *dev, unsigned int flags);
/**@}*/

/**
 * @defgroup conf Configuration
 * Device configuration functions
 * @{
 */

/**
 * It sets the bit mask for the given option
 * @param[in] conf configuration structure where apply this operation
 * @param[in] conf_index configuration value to set (see configuration enum)
 */
static inline void adc_set_conf_mask(struct adc_conf *conf,
				        unsigned int conf_index)
{
	conf->mask |= (1LL << conf_index);
}

/**
 * It assigns a configuration item, and its mask
 * @param[in] conf configuration structure where apply this operation
 * @param[in] conf_index configuration value to set (see configuration enum)
 * @param[in] val value to set
 */
static inline void adc_set_conf(struct adc_conf *conf,
				   unsigned int conf_index, uint32_t val)
{
	conf->value[conf_index] = val;
	adc_set_conf_mask(conf, conf_index);
}



/**
 * It retrieve a configuration value from the configuration descriptor
 * @param[in] conf configuration structure where apply this operation
 * @param[in] conf_index configuration value to set (see configuration enum)
 * @param[out] val value read
 * @return 0 in success, -1 on error
 */
static inline int adc_get_conf(struct adc_conf *conf,
				  unsigned int conf_index,
				  uint32_t *val)
{
	if (conf->mask & (1LL << conf_index)) {
		*val = conf->value[conf_index];
		return 0;
	} else {
		return -1;
	}
}

extern int adc_reset_conf(struct adc_dev *dev, unsigned int flags,
			       struct adc_conf *conf);
extern int adc_apply_config(struct adc_dev *dev, unsigned int flags,
			       struct adc_conf *conf);
extern int adc_retrieve_config(struct adc_dev *dev,
				 struct adc_conf *conf);
extern uint64_t adc_get_capabilities(struct adc_dev *dev,
				     enum adc_configuration_type type);
extern int adc_get_param(struct adc_dev *dev, char *name,
			    char *sptr, int *iptr);
extern int adc_set_param(struct adc_dev *dev, char *name,
			    char *sptr, int *iptr);

/**
 * It enables the configuration for all the supported field by the given device
 * At least the `type` must be set in `conf`
 * @param[in] conf configuration structure to use
 * @param[in] dev the ADC device token
 */
static inline void adc_set_conf_mask_all(struct adc_conf *conf, struct adc_dev *dev)
{
	conf->mask = adc_get_capabilities(dev, conf->type);
}
/**@}*/

/**
 * @defgroup buf Buffer
 * Data management functions
 * @{
 */
extern struct adc_buffer *adc_request_buffer(struct adc_dev *dev,
					     int nsamples,
					     void *(*alloc_fn)(size_t),
					     unsigned int flags);
extern int adc_fill_buffer(struct adc_dev *dev,
			   struct adc_buffer *buf,
			   unsigned int flags,
			   struct timeval *timeout);
extern struct adc_timestamp *adc_tstamp_buffer(struct adc_buffer *buf,
					       struct adc_timestamp *);
extern int adc_release_buffer(struct adc_dev *dev,
			      struct adc_buffer *buf,
			      void (*free_fn)(void *));
/**@}*/

/* libfmcadc version string */
extern const char * const libadc_version_s;

/* zio version string used during compilation of libadc */
extern const char * const libadc_zio_version_s;

/* fmc-adc-100m14b4cha driver version string used during compilation of libadc */
extern const char * const libadc_fmcadc100m14b4cha_version_s;


#ifdef __cplusplus
}
#endif

#endif /* ADC_LIB_H_ */
