/*
 * Initializing and cleaning up the adc library
 *
 * Copyright (C) 2013 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */
#include <string.h>
#include "adc-lib.h"
#include "adc-lib-int.h"

/* * * * * * * * * * * * * * * * Utilities * * * * * * * * * * * * * * * * */

/**
 * This the table of the error messages
 */
static struct adc_errors {
	int num; /**< the errno that linked to the message */
	char *str; /**< the message associated to the errno */
} adc_errors[] = {
	{ ADC_ENOP,		"Operation not supported"},
	{ ADC_ENOCAP,	"Capabilities not supported"},
	{ ADC_ENOCFG,	"Configuration type not supported"},
	{ ADC_ENOGET,	"Cannot get capabilities information"},
	{ ADC_ENOSET,	"Cannot set capabilities information"},
	{ ADC_ENOCHAN,	"Invalid channel"},
	{ ADC_ENOMASK,	"Missing configuration mask"},
	{ ADC_EDISABLED,	"Trigger is disabled: I/O aborted"},
	{ ADC_EROUTE,	"Cannot route correctly the configuration"},
	{ 0, }
};


/**
 * It returns a string describing an error number
 * @param[in] errnum the error number to describe
 * @return a string representing a message error. The function returns static
 *         storage, so you can keep around the pointer it returns, if needed.
 *         Similarly, there is no concurrency between getting the string and
 *         using it, not even in multithreaded environments (``errno`` itself
 *         is already protected, in this respect).
 */
char *adc_strerror(int errnum)
{
	struct adc_errors *p;

	if (errnum < __ADC_ERRNO_START)
		return strerror(errnum);
	for (p = adc_errors; p->num; p++)
		if (p->num == errnum)
			return p->str;
	return "Unknown error code";
}


/**
 * adc_get_driver_type
 * @param[in] dev device which want to know the driver type
 * @return It returns a string that identifies the device type
 */
char *adc_get_driver_type(struct adc_dev *dev)
{
	struct adc_gid *b = (void *)dev;

	return b->board->driver_type;
}


/**
 * It gets the capabilities mask. At least the `type` must be set in `conf`
 * @param[in] dev the ADC device token
 * @param[in] type the configuration type
 * @return It returns capabilities mask for the given configuration type
 */
uint64_t adc_get_capabilities(struct adc_dev *dev,
			      enum adc_configuration_type type)
{
	struct adc_gid *b = (void *)dev;

	if (type >= __ADC_CONF_TYPE_LAST_INDEX)
		return 0;

	return b->board->capabilities[type];
}
