/* Copyright 2013 CERN
 * Author: Federico Vaga <federico.vaga@vaga.pv.it>
 * License: GPLv2
 *
 * This is a simple program to configure the FMC ADC trigger. It is not bug
 * aware because it is only a demo program
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>
#include <pthread.h>
#include <time.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <inttypes.h>
#include <adc-lib.h>

static int arg_show_config = 0;
static int arg_no_read = 0;
static int arg_plot = 0;
static int arg_x_display = 0;

static char git_version[] = "version: " GIT_VERSION;

static void fald_help()
{
	int i;

	printf("\nfald-acq [OPTIONS] -D <device-name>@0x<device-id>\n\n");
	printf(" <device-name>: name of the device type to open\n");
	for (i = 0; i < __ADC_SUPPORTED_BOARDS_LAST_INDEX; i++)
		printf("    %s\n", adc_board_name[i]);
	printf("  --device|-D <device-name>@<device-id>: unique device identifier  (e.g.: \"fmc-adc@0x0400\")\n");
	printf("  --acquisition|-a <parameters>        configure the acquisition\n");
	printf("        <pre-sample>,<post-sample>,<n-shots>,<undersample>\n");
	printf("  --trg-ext <parameters>      configure an external trigger\n");
	printf("        <idx>,<enable>,<polarity>,<delay>\n");
	printf("  --trg-thr <parameters>      configure a threshold trigger\n");
	printf("        <idx>,<enable>,<polarity>,<threshold>,<hysteresis>,<delay>\n");
	printf("  --trg-tim <parameters>      configure a timing trigger\n");
	printf("        <NOT IMPLEMENTED YET>\n");
	printf("  --channel| -c <parameters>  configure an acquisition channel\n");
	printf("        <channel>,<termination>,<range>,<offset>,<saturation>\n");
	printf("  --timeout|-T <millisec>  timeout for acquisition\n");

	printf("  --binary|-B <file>       save binary to <file>\n");
	printf("  --multi-binary|-M <file> save two files per shot: "
						"<file>.0000.ctrl etc\n");
	printf("  --dont-read|-N           config-only\n");
	printf("  --loop|-l <num>          number of loop before exiting\n");
	printf("  --show-data|-s <num>     how many data to display: "
						">0 from head, <0 from tail\n");
	printf("  --graph|-g <chnum>       plot the desired channel\n");
	printf("  --X11|-X                 Gnuplot will use X connection\n");
	printf("  --version|-V             print version information\n");
	printf("  --help|-h                show this help\n\n");
}


enum fald_acq_options {
	FALD_ACQ_OPT_TRG_EXT = 1024,
	FALD_ACQ_OPT_TRG_THR,
	FALD_ACQ_OPT_TRG_TIM,
};

static struct option options[] = {
	{"device",	required_argument, 0, 'D'},
	{"acquisition", required_argument, 0, 'a'},
	{"trg-ext",	required_argument, 0, FALD_ACQ_OPT_TRG_EXT},
	{"trg-thr",	required_argument, 0, FALD_ACQ_OPT_TRG_THR},
	{"trg-tim",	required_argument, 0, FALD_ACQ_OPT_TRG_TIM},
	{"channel",	required_argument, 0, 'c'},
	{"timeout",	required_argument, 0, 'T'},

	/* new options, to help stress-test */
	{"binary",	required_argument, 0, 'B'},
	{"multi-binary",required_argument, 0, 'M'},
	{"dont-read",	no_argument,       &arg_no_read, 1},
	{"loop",	required_argument, 0, 'l'},
	{"show-data",	required_argument, 0, 's'},
	{"plot",	no_argument, &arg_plot, 1},
	{"X11", no_argument, &arg_x_display , 1},

	/* loop for stess test */
	{"loop",	required_argument, 0, 'l'},

	{"version",	no_argument,       0, 'V'},
	{"help", 	no_argument,       0, 'h'},
	{"show-config", no_argument,       &arg_show_config, 1},
	{0, 0, 0, 0}
};

#define GETOPT_STRING "D:Vha:c:T:B:M:l:s:g:"

static void print_version(char *pname)
{
	printf("%s %s\n", pname, git_version);
	adc_print_version();
}

/* variables shared between threads */
static char *devname = NULL;
static unsigned int devid = 0;
static int show_ndata = INT_MAX; /* by default all values are displayed */
static int binmode;
static int timeout = -1;
static int loop = 1;
static char *basefile;
static unsigned int nsamples = 0;
static unsigned int ssize = 0;
static char *_argv[16];
static uint32_t nchan;

/* default is 1 V*/
static double bit_scale = 0.5/(1<<15);



/**
 * It shows the current configuration
 * @param[in] adc the ADC token
 */
static void fald_acq_print_config(struct adc_dev *adc)
{
	struct adc_conf cfg, cfg_brd;
	uint32_t value, n;
	int err, i;

	if (!arg_show_config)
		return;
	/*
	 * Show the configuration only the first time,
	 * not at every loop iteration
	 */
	arg_show_config = 0;

	fprintf(stdout, "Board Configuration\n");
	memset(&cfg_brd, 0, sizeof(struct adc_conf));
	cfg_brd.type = ADC_CONF_TYPE_BRD;
	adc_set_conf_mask_all(&cfg_brd, adc);
	err = adc_retrieve_config(adc, &cfg_brd);
	if (err) {
		fprintf(stderr,
			"Cannot retrieve the board configuration: (%d) %s\n",
			errno, adc_strerror(errno));
	} else {
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_STATUS, &value);
		if (!err)
			fprintf(stdout, "\tStatus: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_STATE_MACHINE_STATUS, &value);
		if (!err)
			fprintf(stdout, "\tState machine: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_MAX_FREQ_HZ, &value);
		if (!err)
			fprintf(stdout, "\tMax sampling frequency: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_MIN_FREQ_HZ, &value);
		if (!err)
			fprintf(stdout, "\tMin sampling frequency: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_CHAN, &value);
		if (!err)
			fprintf(stdout, "\tNumber of channels: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_EXT, &value);
		if (!err)
			fprintf(stdout, "\tNumber of external triggers: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_THR, &value);
		if (!err)
			fprintf(stdout, "\tNumber of threshold triggers: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_TIM, &value);
		if (!err)
			fprintf(stdout, "\tNumber of timer triggers: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_UTC_TIMING_BASE_S, &value);
		if (!err)
			fprintf(stdout, "\tTime Seconds: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_UTC_TIMING_BASE_T, &value);
		if (!err)
			fprintf(stdout, "\tTime sub-second ticks: %d\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_UTC_TIMING_BASE_B, &value);
		if (!err)
			fprintf(stdout, "\tTime sub-second bins: %d\n", value);
	}

	fprintf(stdout, "Acquisition Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_ACQ;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_retrieve_config(adc, &cfg);
	if (err) {
		fprintf(stderr,
			"Cannot retrieve the acquisition configuration: (%d) %s\n",
			errno, adc_strerror(errno));
	} else {
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_N_SHOTS, &value);
		if (!err)
			fprintf(stdout, "\tShots: %d\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_PRE_SAMP, &value);
		if (!err)
			fprintf(stdout, "\tPre samples: %d\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_POST_SAMP, &value);
		if (!err)
			fprintf(stdout, "\tPost samples: %d\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_UNDERSAMPLE, &value);
		if (!err)
			fprintf(stdout, "\tUndersample (decimation): %d\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_FREQ_HZ, &value);
		if (!err)
			fprintf(stdout, "\tSampling rate: %d Hz\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_N_BITS, &value);
		if (!err)
			fprintf(stdout, "\tBit per sample: %d\n", value);
	}

	fprintf(stdout, "Channel Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_CHN;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_CHAN, &n);
	if (err)
		n = 0;
	for (i = 0; i < n; ++i) {
		fprintf(stdout, "\tChannel %d\n", i + 1);
		cfg.route_to = i;
		err = adc_retrieve_config(adc, &cfg);
		if (err) {
			fprintf(stderr,
				"Cannot retrieve the external trigger configuration (): (%d) %s\n",
				errno, adc_strerror(errno));
		} else {
			err = adc_get_conf(&cfg, ADC_CONF_CHN_RANGE, &value);
			if (!err)
				fprintf(stdout, "\t\tRange: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_CHN_OFFSET, &value);
			if (!err)
				fprintf(stdout, "\t\tOffset: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_CHN_SATURATION, &value);
			if (!err)
				fprintf(stdout, "\t\tSaturation: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_CHN_TERMINATION, &value);
			if (!err)
				fprintf(stdout, "\t\tTermination: %d\n", value);
		}
	}

	fprintf(stdout, "Trigger (External) Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_EXT;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_EXT, &n);
	if (err)
		n = 0;
	for (i = 0; i < n; ++i) {
		fprintf(stdout, "\tExternal trigger %d\n", i + 1);
		cfg.route_to = i;
		err = adc_retrieve_config(adc, &cfg);
		if (err) {
			fprintf(stderr,
				"Cannot retrieve the external trigger configuration (): (%d) %s\n",
				errno, adc_strerror(errno));
		} else {
			err = adc_get_conf(&cfg, ADC_CONF_TRG_EXT_ENABLE, &value);
			if (!err)
				fprintf(stdout, "\t\tEnable: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_EXT_POLARITY, &value);
			if (!err)
				fprintf(stdout, "\t\tPolarity: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_EXT_DELAY, &value);
			if (!err)
				fprintf(stdout, "\t\tDelay: %d\n", value);
		}
	}

	fprintf(stdout, "Trigger (Threshold) Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_THR;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_THR, &n);
	if (err)
		n = 0;
	for (i = 0; i < n; ++i) {
		fprintf(stdout, "\tThreshold trigger %d\n", i + 1);
		cfg.route_to = i;
		err = adc_retrieve_config(adc, &cfg);
		if (err) {
			fprintf(stderr,
				"Cannot retrieve the external trigger configuration (): (%d) %s\n",
				errno, adc_strerror(errno));
		} else {
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_ENABLE, &value);
			if (!err)
				fprintf(stdout, "\t\tEnable: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_POLARITY, &value);
			if (!err)
				fprintf(stdout, "\t\tPolarity: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_DELAY, &value);
			if (!err)
				fprintf(stdout, "\t\tDelay: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_THRESHOLD, &value);
			if (!err)
				fprintf(stdout, "\t\tThreshold: %d\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_HYSTERESIS, &value);
			if (!err)
				fprintf(stdout, "\t\tHysteresis: %d\n", value);
		}
	}

	fprintf(stdout, "Trigger (Time) Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_EXT;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_retrieve_config(adc, &cfg);
	if (err)
		fprintf(stderr,
			"Cannot retrieve the time trigger configuration: (%d) %s\n",
			errno, adc_strerror(errno));
}


/**
 * It writes data on file
 * @param[in] filename file where write data
 * @param[in] ch the channel data to write
 * @param[in] pdata data from ADC (is interleaved)
 * @param[in] n_sample number of sample in pdata (is interleaved)
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int write_file(const char *filename, int ch, int16_t *pdata,
		      unsigned int n_sample)
{
	FILE *filp;
	int i;
	double val;

	int mask = umask(0); /* set umask to have a file with 0666 mode */
	filp = fopen(filename, "w");
	umask(mask); /* restore previous umask */
	if (filp == NULL) {
		printf("fopen %s failed: %s", filename, strerror(errno));
		return -1;
	}
	for (i = 0, pdata += (ch - 1); i < n_sample; ++i, pdata += 4) {
		val = (*pdata)*bit_scale;
		fprintf(filp, "%d %g\n", i, val);
	}
	if (fclose(filp) < 0)
		return -1;
	return 0;
}


/**
 * It configures the acquisition according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<pre-sample>,<post-sample>,<n-shots>,<undersample>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_acq_acquisition_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t pre, post, nshot, under;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &pre, &post, &nshot, &under);
	if (ret < 0) {
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_ACQ;

	switch (ret) {
	case 4:
		adc_set_conf(&cfg, ADC_CONF_ACQ_UNDERSAMPLE, under);
	case 3:
		adc_set_conf(&cfg, ADC_CONF_ACQ_N_SHOTS, nshot);
	case 2:
		adc_set_conf(&cfg, ADC_CONF_ACQ_POST_SAMP, post);
	case 1:
		adc_set_conf(&cfg, ADC_CONF_ACQ_PRE_SAMP, pre);
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}


/**
 * It configures a channel according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<channel>,<termination>,<range>,<offset>,<saturation>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_acq_channel_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t ch, term, range, offset, sat;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &ch, &term, &range, &offset, &sat);
	if (ret <= 0) { /* channel is mandatory */
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_CHN;
	cfg.route_to = ch;

	switch (ret) {
	case 5:
		adc_set_conf(&cfg, ADC_CONF_CHN_SATURATION, sat);
	case 4:
		adc_set_conf(&cfg, ADC_CONF_CHN_OFFSET, offset);
	case 3:
		/*
		 * in-range
		 * 0x23 (35): 100mV range
		 * 0x11 (17): 1V range
		 * 0x45 (69): 10V range
		 * 0x00 (0): Open input
		 */
		switch (range) {
		case 100:
			range = 0x23;
			bit_scale = 0.05/(1<<15);
			break;
		case 1:
			range = 0x11;
			bit_scale = 0.5/(1<<15);
			break;
		case 10:
			range = 0x45;
			bit_scale = 5.0/(1<<15);
			break;
		}
		adc_set_conf(&cfg, ADC_CONF_CHN_RANGE, range);
	case 2:
		adc_set_conf(&cfg, ADC_CONF_CHN_TERMINATION, term);
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}


/**
 * It configures a external trigger according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<idx>,<enable>,<polarity>,<delay>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_trg_external_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t n, enable, polarity, delay;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &n, &enable, &polarity, &delay);
	if (ret <= 0) { /* channel is mandatory */
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_EXT;
	cfg.route_to = n;

	switch (ret) {
	case 4:
		adc_set_conf(&cfg, ADC_CONF_TRG_EXT_DELAY, delay);
	case 3:
		adc_set_conf(&cfg, ADC_CONF_TRG_EXT_POLARITY, polarity);
	case 2:
		adc_set_conf(&cfg, ADC_CONF_TRG_EXT_ENABLE, enable);
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}


/**
 * It configures a threshold trigger according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<idx>,<enable>,<polarity>,<threshold>,<hysteresis>,<delay>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_trg_threshold_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t n, enable, polarity, delay, threshold, hysteresis;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &n, &enable, &polarity, &threshold, &hysteresis, &delay);
	if (ret <= 0) { /* channel is mandatory */
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_THR;
	cfg.route_to = n;

	switch (ret) {
	case 6:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_DELAY, delay);
	case 5:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_HYSTERESIS, hysteresis);
	case 4:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_THRESHOLD, threshold);
	case 3:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_POLARITY, polarity);
	case 2:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_ENABLE, enable);
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}


/**
 * It configures a timer trigger according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<channel>,<termination>,<range>,<offset>,<saturation>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_trg_timer_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_TIM;
	cfg.route_to = 0;

	return adc_apply_config(adc, 0 , &cfg);
}


/**
 * It parses command line arguments and consequential it configures the device
 * @param[in] argc number of arguments
 * @param[in] argv arguments
 */
static int fald_acq_parse_args_and_configure(struct adc_dev *adc, int argc, char *argv[])
{
	int c, opt_index;
	int err;

	optind = 1; /* set to 1 to make getopt_long happy */
	/* Parse options */
	while ((c = getopt_long(argc, argv, GETOPT_STRING, options, &opt_index))
	       >= 0 ) {
		switch (c) {
		case 'a':
			err = fald_acq_acquisition_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure acquisition: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case 'c':
			err = fald_acq_channel_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure channel: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case FALD_ACQ_OPT_TRG_EXT:
			err = fald_trg_external_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure external trigger: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case FALD_ACQ_OPT_TRG_THR:
			err = fald_trg_threshold_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure threshold trigger: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case FALD_ACQ_OPT_TRG_TIM:
			err = fald_trg_timer_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure timer trigger: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case 'T':
			timeout = atoi(optarg);
			break;
		case 'B':
			binmode = 1; /* do binary (default is 0) */
			basefile = optarg;
			break;
		case 'M':
			binmode = 2; /* do many binaries */
			basefile = optarg;
			break;
		case 'l':
			loop = atoi(optarg);
			break;
		case 's':
			show_ndata = atoi(optarg);
			break;
		default:
			break;
		}
	}

	fald_acq_print_config(adc);

	return 0;
}


/**
 * It prints data to stdout
 * @param[in] ctrl
 * @param[in] buf buffer to print
 * @param[in] acq_cfg acquisition configuration associated to the buffer
 */
static void fald_acq_print_data(struct adc_buffer *buf,
				struct adc_conf *acq_cfg,
				unsigned int n)
{
	int j, ch;
	int16_t *data;  /* FMC-ADC-100M sample size is 14bit, 16bit for ZIO */
	uint32_t pre;

	if (n == 0)
		return;

	if (nsamples != buf->nsamples) {
		fprintf(stdout, "discrepancy between nsamples: %d and buf->nsamples: %d\n",
			nsamples, buf->nsamples);
		return;
	}

	adc_get_conf(acq_cfg, ADC_CONF_ACQ_PRE_SAMP, &pre);

	data = buf->data;
	/* Print data */
	for (j = 0; j < nsamples; j++) {
		if ( (n > 0 && j < n) ||
		     (n < 0 && (nsamples - j) <= (-n)) ) {
			printf("%5i     ", j - pre);
			for (ch = 0; ch < nchan; ch++)
				printf("%7i", *(data++));
			printf("\n");
		} else {
			data += nchan;
		}
	}
}

/**
 * Save data to a single file
 * @param[in] buf buffer to store
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int fald_acq_write_single(struct adc_buffer *buf)
{
	char fname[PATH_MAX];
	int16_t *data;  /* FMC-ADC-100M sample size is 14bit, 16bit for ZIO */
	FILE *f;
	int err = 0;

	sprintf(fname, "%s", basefile);
	f = fopen(fname, "a");
	if (!f) {
		fprintf(stderr, "%s: %s\n",
			fname, strerror(errno));
		exit(1);
	}
#ifdef FIXME
	ctrl = buf->metadata;
	if (fwrite(ctrl, sizeof(struct zio_control), 1, f) != 1)
		err++;
#endif
	data = buf->data;
	if (fwrite(data, ssize, nsamples, f) != nsamples)
		err++;

	if (err) {
		fprintf(stderr, "write(%s): short write\n",
			basefile);
		return -1;
	}

	return 0;
}


/**
 * Sava data to different file for control and data for each shot
 * @param[in] buf buffer to store
 * @param[in] shot_i i-th shot acquisition
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int fald_acq_write_multiple(struct adc_buffer *buf,
				   unsigned int shot_i)
{
	char fname[PATH_MAX];
	uint16_t *data;
	FILE *f;
#ifdef FIXME
	/* FIXME zio control is specific. The library should provide
	   at least the metadata size so that automatic tools can at least
	   read */
	sprintf(fname, "%s.%03i.ctrl", basefile, shot_i);
	f = fopen(fname, "w");
	if (!f) {
		fprintf(stderr, "%s: %s\n",
			fname, strerror(errno));
		return -1;
	}
	ctrl = buf->metadata;
	if (fwrite(ctrl, sizeof(struct zio_control), 1, f) != 1) {
		fprintf(stderr, "write(%s): short write\n", fname);
		return -1;
	}
	fclose(f);
#endif
	sprintf(fname, "%s.%03i.data", basefile, shot_i);
	f = fopen(fname, "w");
	if (!f) {
		fprintf(stderr, "%s: %s\n",
			fname, strerror(errno));
		return -1;
	}
	data = buf->data;
	if (fwrite(data, ssize, nsamples, f)
	    != nsamples) {
		fprintf(stderr, "write(%s): short write\n", fname);
		return -1;
	}
	fclose(f);
	return 0;
}


/**
 * Plot data using gnuplot
 * @param[in] buf buffer to plot
 * @param[in] ch channel to show
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static void fald_acq_plot_data(struct adc_buffer *buf, unsigned int ch)
{
	int16_t *data = buf->data;
	char fname[PATH_MAX];
	char cmd[256];

	snprintf(fname, sizeof(fname), "/tmp/fmcadc.0x%04x.ch%d.dat", devid, ch);
	if (write_file(fname, ch, data, (nsamples)/4) < 0) {
		printf("Cannot plot data. Write data into file %s failed.\n", fname);
		return;
	}
	snprintf(cmd, sizeof(cmd), "echo \"%s plot '%s' with lines "
		 "\" | gnuplot -persist",
		 (arg_x_display == 1 ? "set term x11; " : "set term dumb; "),
		 fname);
	cmd[sizeof(cmd) - 1] = '\0';
	system(cmd);
}


/**
 * It process the buffer
 * @param[in] buf the buffer to use
 * @param[in] iteration loop iteration number
 */
static void fald_acq_process_buffer(struct adc_buffer *buf,
				    struct adc_conf *cfg_acq,
				    unsigned int iteration)
{
	int w;

	/* print/store data*/
	switch(binmode) {
	default:
		fald_acq_print_data(buf, cfg_acq, show_ndata);
		break;
	case 1:
		fald_acq_write_single(buf);
		break;
	case 2:
		fald_acq_write_multiple(buf, iteration);
		break;
	}

	if (arg_plot)
		for (w = 0; w < nchan; ++w)
			fald_acq_plot_data(buf, w + 1);
}

/**
 * It parses the command line arguments for the basic parameters
 * @param[in] argc number of arguments
 * @param[in] argv arguments
 * @return It returns 0 on success, otherwise -1 and errno is appropriately set
 *
 * It looks for:
 * - device ID
 * - help
 * - version
 */
static int fald_acq_parse_args_basic(int argc, char *argv[])
{
	int opt_index;
	char c;

	while ((c = getopt_long(argc, argv, GETOPT_STRING,
				options, &opt_index)) >= 0 ) {
		switch (c) {
		case 'D': {
			char *strdevid;
			int len;

			strdevid = strchr(optarg, '@');
			if (!strdevid) {
				fprintf(stderr, "%s: invalid device \"%s\" (<type>@<device-id>)\n",
					_argv[0], optarg);
				exit(1);
			}
			strdevid++; /* +1 to skip '@' */
			sscanf(strdevid, "0x%x", &devid);

			/* -1 count '@' it will be replaced by 0
			 * +1 for \0
			 */
			len = (int)(strdevid - 1 - optarg + 1);
			devname = malloc(len);
			if (!devname) {
				fprintf(stderr, "Cannot allocate memory\n");
				exit(1);
			}
			memset(devname, 0, len);
			strncpy(devname, optarg, len - 1);

			fprintf(stdout, "%s:  %s -> type:%s, devid:%s\n",
				_argv[0], optarg, devname, strdevid);
			break;
		}
		case 'V':
			print_version(argv[0]);
			exit(EXIT_SUCCESS);
			break;
		case 'h':
			fald_help();
			exit(EXIT_SUCCESS);
			break;
		default:
			break;
		}
	}
	return 0;
}

int main(int argc, char *argv[])
{
	struct adc_dev *adc;
	struct adc_buffer *buf;
	struct adc_timestamp ts;
	struct timeval tv = {0, 0};
	struct adc_conf cfg_brd, cfg_acq;
	uint32_t pre, post, bits, nshots;
	int i, k, err;


	err = adc_init();
	if (err)
		exit(EXIT_FAILURE);

	/* set local _argv[0] with  pg name */
	_argv[0] = argv[0];
	err = fald_acq_parse_args_basic(argc, argv);
	if (err)
		exit(err);

	if (!devname) {
		fprintf(stderr, "%s: device is a mandatory argument (%s@0x%x)\n",
			argv[0], devname, devid);
		fald_help();
		goto err_devid_invalid;
	}

	/* Open the ADC */
	adc = adc_open(devname, devid, 0, 0, ADC_F_FLUSH);
	if (!adc) {
		fprintf(stderr, "%s: cannot open device: %s\n",
			argv[0], adc_strerror(errno));
		goto err_open;
	}

	/* get the new given trigger and acq config
	   Only the ones provided will override the current ones */
	err = fald_acq_parse_args_and_configure(adc, argc, argv);
	if (err)
		goto err_parse;

	/*
	 * We already know the configuration, but here we will retrieve it
	 * from the library anyway as an example on how to get the
	 * configuration
	 */
	memset(&cfg_brd, 0, sizeof(struct adc_conf));
	cfg_brd.type = ADC_CONF_TYPE_BRD;
	adc_set_conf_mask_all(&cfg_brd, adc);
	err = adc_retrieve_config(adc, &cfg_brd);
	if (err) {
		fprintf(stderr,
			"Cannot retrive the board configuration: (%d) %s\n",
			errno, adc_strerror(errno));
		goto err_conf_get;
	}
	memset(&cfg_acq, 0, sizeof(struct adc_conf));
	cfg_acq.type = ADC_CONF_TYPE_ACQ;
	adc_set_conf_mask_all(&cfg_acq, adc);
	err = adc_retrieve_config(adc, &cfg_acq);
	if (err) {
		fprintf(stderr,
			"Cannot retrive the acquisition configuration: (%d) %s\n",
			errno, adc_strerror(errno));
		goto err_conf_get;
	}

	adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_CHAN, &nchan);
	adc_get_conf(&cfg_acq, ADC_CONF_ACQ_PRE_SAMP, &pre);
	adc_get_conf(&cfg_acq, ADC_CONF_ACQ_POST_SAMP, &post);
	adc_get_conf(&cfg_acq, ADC_CONF_ACQ_N_BITS, &bits);
	adc_get_conf(&cfg_acq, ADC_CONF_ACQ_N_SHOTS, &nshots);
	nsamples = pre + post;
	ssize = (bits + (8 - 1)) / 8; 	/* get sample size - round up */

	buf = adc_request_buffer(adc, nsamples, NULL /* alloc */, 0);
	if (!buf) {
		fprintf(stderr, "Cannot allocate buffer (%s)\n",
			adc_strerror(errno));
		exit(1);
	}

	for (i = 0; i < loop && arg_no_read == 0; ++i) {
		fprintf(stderr, "Acquisition %i\n", i + 1);
		/* Stop any pending acquisition */
		err = adc_acq_stop(adc, 0);
		if (err) {
			fprintf(stderr, "Cannot stop pending acquisition: (%d) %s\n",
				errno, adc_strerror(errno));
			continue;
		}

		err = adc_acq_start(adc, ADC_F_FLUSH, &tv);
		if (err) {
			fprintf(stderr, "Cannot start acquisition: (%d) %s\n",
				errno, adc_strerror(errno));
			continue;
		}
		for (k = 0; k < nshots; ++k) {
			fprintf(stderr,
				"iteration: %d, shot: %i/%i, nsamples: %d\n",
				i, k + 1, nshots, nsamples);

			err = adc_acq_poll(adc, 0 , NULL);
			if (err) {
				fprintf(stderr, "Failed to wait for data: (%d) %s\n",
					errno, adc_strerror(errno));
				break;
			}

			err = adc_fill_buffer(adc, buf, 0, NULL);
			if (err) {
				fprintf(stderr, "Failed to retrieve for data: (%d) %s\n",
					errno, adc_strerror(errno));
				break;
			}
			adc_tstamp_buffer(buf, &ts);
			fprintf(stderr,
				"time: {secs:%"PRIu64" ticks:%"PRIu64" bins:%"PRIu64"}\n",
				ts.secs, ts.ticks, ts.bins);

			fald_acq_process_buffer(buf, &cfg_acq, i);
		}
	}

	adc_release_buffer(adc, buf, NULL);

	adc_close(adc);
	adc_exit();
	free(devname);
	exit(EXIT_SUCCESS);

err_conf_get:
err_parse:
	adc_close(adc);
err_open:
err_devid_invalid:
	adc_exit();
	free(devname);
	exit(EXIT_FAILURE);
}
