/* Copyright 2013 CERN
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * License: GPLv2
 *
 * This is an example program that shows the different part of the library
 * in action.
 *
 *
 * This is part of the documentation, so when you change it **REMEMBER** to
 * update the references in documents in `doc/`.
 *
 * The reason to have the example here as a true program is that we can always
 * compile it and run it, so that we can validate the example.
 * This way, at least for this aspect, the documentation in always aligned.
 *
 * Do not try to optimize this code, the purpose is to show code, highlight
 * snippet and being able to refer to it from the documentation
 */
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <adc-lib.h>

#define NSHOT 3
#define PRESAMPLES 5
#define POSTSAMPLES 5

static char *prog_name;

static int current_config_acquisition(struct adc_dev *adc);
static void process_buffer(struct adc_buffer *buf);
static int config_acquisition(struct adc_dev *adc);
static int config_and_acquire(struct adc_dev *adc, struct adc_buffer *buf);
static int alloc_config_and_acquire(struct adc_dev *adc);
static int use_adc_library();

int main(int argc, char *argv[])
{
	int err;

	prog_name = argv[0];

	err = adc_init();
	if (err)
		exit(EXIT_FAILURE);

	err = use_adc_library();
	if (err)
		exit(EXIT_FAILURE);
	adc_exit(); /* optional, indeed in the error condition
		       we do not do it */
	exit(EXIT_SUCCESS);
}

static int use_adc_library()
{
	struct adc_dev *adc;
	int err;

	/* Open the ADC */
	adc = adc_open("adc-genericfake", 0x0000, 0, 0, ADC_F_FLUSH);
	if (!adc) {
		fprintf(stderr, "%s: Cannot open device: %s\n",
			prog_name, adc_strerror(errno));
		return -1;
	}

	err = alloc_config_and_acquire(adc);

	adc_close(adc);

	return err;
}

static int alloc_config_and_acquire(struct adc_dev *adc)
{
	struct adc_buffer *buf;
	int err;

	buf = adc_request_buffer(adc, NSHOT * (PRESAMPLES + POSTSAMPLES),
				 NULL, 0);
	if (!buf) {
		fprintf(stderr, "%s: Cannot allocate buffer: (%d) %s\n",
			prog_name, errno, adc_strerror(errno));
		return -1;
	}

	err = config_and_acquire(adc, buf);

	adc_release_buffer(adc, buf, NULL);

	return err;
}


static int config_and_acquire(struct adc_dev *adc, struct adc_buffer *buf)
{
	struct timeval tv = {0, 0};
	int err, k;

	err = config_acquisition(adc);
	if (err)
		return err;

	/* stop any pending acquisition on the device */
	err = adc_acq_stop(adc, 0);
	if (err) {
		fprintf(stderr,
			"%s: cannot stop pending acquisition: (%d) %s\n",
			prog_name, errno, adc_strerror(errno));
		return err;
	}
	/* Start the acquisition without wait */
	err = adc_acq_start(adc, ADC_F_FLUSH, &tv);
	if (err) {
		fprintf(stderr,
			"%s: Cannot start acquisition: (%d) %s\n",
			prog_name, errno, adc_strerror(errno));
		return err;
	}

	for (k = 0; k < NSHOT; ++k) {
		fprintf(stdout, "Shot %d/%d\n", k + 1, NSHOT);
		/* Wait until the data is available */
		err = adc_acq_poll(adc, 0 , NULL);
		if (err) {
			fprintf(stderr, "Failed to wait for data: (%d) %s\n",
				errno, adc_strerror(errno));
			break;
		}
		/* Fill the buffer with the data */
		err = adc_fill_buffer(adc, buf, 0, NULL);
		if (err) {
			fprintf(stderr, "Failed to retrieve for data: (%d) %s\n",
				errno, adc_strerror(errno));
			break;
		}
		process_buffer(buf);
	}

	/* stop the acquisition on the device */
	err = adc_acq_stop(adc, 0);
	if (err) {
		fprintf(stderr,
			"%s: cannot stop the acquisition: (%d) %s\n",
			prog_name, errno, adc_strerror(errno));
		return err;
	}

	return 0;
}

static int config_acquisition(struct adc_dev *adc)
{
	struct adc_conf cfg;
	int err;

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_ACQ;
	adc_set_conf(&cfg, ADC_CONF_ACQ_N_SHOTS, NSHOT);
	adc_set_conf(&cfg, ADC_CONF_ACQ_POST_SAMP, POSTSAMPLES);
	adc_set_conf(&cfg, ADC_CONF_ACQ_PRE_SAMP, PRESAMPLES);
	err = adc_apply_config(adc, 0 , &cfg);
	if (err) {
		fprintf(stderr, "%s: Cannot configure acquisition: (%d) %s\n",
			prog_name, errno, adc_strerror(errno));
		return err;
	}

	/* read back the configuration */
	current_config_acquisition(adc);

	return 0;
}

static int current_config_acquisition(struct adc_dev *adc)
{
	struct adc_conf cfg;
	uint32_t sh, pr, po;
	int err;

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_ACQ;
	adc_set_conf_mask(&cfg, ADC_CONF_ACQ_N_SHOTS);
	adc_set_conf_mask(&cfg, ADC_CONF_ACQ_POST_SAMP);
	adc_set_conf_mask(&cfg, ADC_CONF_ACQ_PRE_SAMP);
	err = adc_retrieve_config(adc, &cfg);
	if (err) {
		fprintf(stderr, "%s: Cannot configure acquisition: (%d) %s\n",
			prog_name, errno, adc_strerror(errno));
		return err;
	}

	adc_get_conf(&cfg, ADC_CONF_ACQ_N_SHOTS, &sh);
	fprintf(stdout, "Number of shots: %d\n", sh);
	adc_get_conf(&cfg, ADC_CONF_ACQ_POST_SAMP, &po);
	fprintf(stdout, "Number of post-samples: %d\n", po);
	adc_get_conf(&cfg, ADC_CONF_ACQ_PRE_SAMP, &pr);
	fprintf(stdout, "Number of pre-samples: %d\n", pr);

	return 0;

}

static void process_buffer(struct adc_buffer *buf)
{
	struct adc_timestamp ts;
	uint8_t *data;
	int i;

	adc_tstamp_buffer(buf, &ts);
	fprintf(stdout,
		"time: {secs:%"PRIu64" ticks:%"PRIu64" bins:%"PRIu64"}\n",
		ts.secs, ts.ticks, ts.bins);
	data = buf->data;
	for (i = -PRESAMPLES; i < POSTSAMPLES; ++i, ++data)
		fprintf(stdout, "sample: %d, value: 0x%02"PRIx8"\n",
			i, *data);
}
